# todo : avoid blocking when can't reach desired value

from machine import Pin, ADC, freq
import time

# Standard QC2.0 are 50 (5V), 90 (9V), 120 (12V) or 200 (20V)
# Any other will work using QC3.0
# TARGET_DCV = 120



CLASSA = 120  # Class A QC is up to 12V
CLASSB = 200  # Class B QC is up to 20V (so far?)

QC_T_HANDSHAKE_S = 2
QC_T_GLICH_V_CHANGE_MS = 60
QC_CONTINUOUS_PULSE_MS = 5

startDcV = 44
maxDcV = CLASSB
minDcV = 36

switch = Pin(26, Pin.IN, Pin.PULL_UP)
DN = Pin(27)
DP = Pin(28)

rotA = Pin(1, Pin.IN, Pin.PULL_UP)
rotB = Pin(2, Pin.IN, Pin.PULL_UP)
rotPush = Pin(4, Pin.IN, Pin.PULL_UP)
button = Pin(3, Pin.IN, Pin.PULL_UP)


sense = ADC(Pin(29))

PIN_SENSE = ADC(Pin(29))  # Pin for reading actual voltage (through 1/11 bridge divider)

freq(48000000)

def setup():
    Init_QC()
    setQCdcV(startDcV)

# D+ pin (resistors bridge controlled)
def dp0mV():
    DP.low()
    DP.init(mode=Pin.OUT)

def dp600mV():
    DP.init(mode=Pin.IN)

def dp3300mV():
    DP.high()
    DP.init(mode=Pin.OUT)

# D- pin (DAC controlled)
def dn0mV():
    DN.low()
    DN.init(mode=Pin.OUT)

def dn600mV():
    DN.init(mode=Pin.IN)

def dn3300mV():
    DN.high()
    DN.init(mode=Pin.OUT)

def dnFloat():
    DN.init(mode=Pin.IN)

def Init_QC():
    # Make D- pin float
    dnFloat()
    # Write 0.6V on D+
    dp600mV()
    # Wait (at least) 2sec
    time.sleep(QC_T_HANDSHAKE_S)
    
    dn0mV()
    
    time.sleep_ms(QC_T_GLICH_V_CHANGE_MS)
    # Now QC protocol is active

def QCVoltageDecrease02():
    dn600mV()
    time.sleep_ms(QC_CONTINUOUS_PULSE_MS)
    dn3300mV()
    time.sleep_ms(QC_CONTINUOUS_PULSE_MS)

def QCVoltageIncrease02():
    dp3300mV()
    time.sleep_ms(QC_CONTINUOUS_PULSE_MS)
    dp600mV()
    time.sleep_ms(QC_CONTINUOUS_PULSE_MS)

def setQCdcV(dcV):
    # See CHY100 datasheet (Table 1 - Output Voltage Lookup Table)
    # https://www.mouser.be/datasheet/2/328/chiphy_family_datasheet_269468-3198404.pdf
    # 5.0V
    if dcV == 50:
        dp600mV()
        dn0mV()
    # 9.0V
    elif dcV == 90:
        dp3300mV()
        dn600mV()
    # 12.0V
    elif dcV == 120:
        dp600mV()
        dn600mV()
    # 20.0V
    elif dcV == 200:
        dp3300mV()
        dn3300mV()
    # Other (via QC3.0 +-200mV steps)
    else:
        dcVNow = 50

        # Speed up rise for dcVs > 120
        if dcV >= 120:
            dp600mV()
            dn600mV()
            dcVNow = 120
            time.sleep_ms(QC_T_GLICH_V_CHANGE_MS)

        # Enter continuous mode
        dn3300mV()
        time.sleep_ms(QC_T_GLICH_V_CHANGE_MS)

        if dcV > dcVNow:
            while dcVNow < dcV:
                QCVoltageIncrease02()
                dcVNow += 2
        else:
            while dcVNow > dcV:
                QCVoltageDecrease02()
                dcVNow -= 2
    time.sleep_ms(QC_T_GLICH_V_CHANGE_MS)

setup()


lastState = (rotA.value(), rotB.value())

curDcVTarget = startDcV

def adjust() :
    global curDcVTarget
    val = sense.read_u16() / 250
    lastVal = val
    errCount = 0
    if val > curDcVTarget : 
        while val > curDcVTarget + 2 :
            QCVoltageDecrease02()
            val = sense.read_u16() / 250
            if lastVal - val < 1 :
                errCount += 1
                if errCount > 3:
                    errCount = 0
                    print("Couldn't go that low. (power supply limit)")
                    break
            lastVal = val
    else :
        while val < curDcVTarget - 2 :
            QCVoltageIncrease02()
            val = sense.read_u16() / 250
            if val - lastVal < 1 :
                errCount += 1
                if errCount > 3:
                    errCount = 0
                    print("Couldn't go that high. (power supply limit)")
                    break
            lastVal = val
    print("measured value : ", round(val/10,1), "V")

def printTarget():
    global curDcVTarget
    print("target value : ", round(curDcVTarget/10,1), "V")

def clockwise():
    global curDcVTarget
    curDcVTarget += 1
    if curDcVTarget > maxDcV :
        curDcVTarget = maxDcV
    if switch.value() == 0 :
        adjust()
    printTarget()

def cClockwise():
    global curDcVTarget
    curDcVTarget -= 1
    if curDcVTarget < minDcV :
        curDcVTarget = minDcV
    if switch.value() == 0 :
        adjust()
    printTarget()
    
def pushAction():
    global curDcVTarget
    adjust()
    printTarget()

transition_table = {
    (0, 0, 0, 1): cClockwise,
    (0, 1, 1, 1): cClockwise,
    (1, 1, 1, 0): cClockwise,
    (1, 0, 0, 0): cClockwise,
    (0, 0, 1, 0): clockwise,
    (1, 0, 1, 1): clockwise,
    (1, 1, 0, 1): clockwise,
    (0, 1, 0, 0): clockwise,
}

while True:

    # check rotary encoder movements
    curState = (rotA.value(), rotB.value())
    transition = (lastState[0], lastState[1], curState[0], curState[1])
    if transition in transition_table:
        transition_table[transition]()

    lastState = curState

    # check rotary push button
    if rotPush.value() == 0 :
        pushAction()
        time.sleep(0.01)
        while rotPush.value() == 0:
            pass
    
    time.sleep(0.01)


